from abc import abstractmethod

from pygfunction.boreholes import Borehole

from ghedesigner.constants import PI, TWO_PI
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil


class GHEDesignerBoreholeBase:
    def __init__(
        self,
        m_flow_borehole: float,
        fluid: Fluid,
        borehole: Borehole,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
    ) -> None:
        self.m_flow_borehole = m_flow_borehole
        self.pipe = pipe
        self.soil = soil
        self.grout = grout
        self.fluid = fluid
        self.borehole = borehole

    @abstractmethod
    def calc_effective_borehole_resistance(self) -> float:
        pass

    @staticmethod
    def compute_fluid_resistance(h_conv: float, radius: float) -> float:
        return 1 / (h_conv * TWO_PI * radius)

    @staticmethod
    def compute_reynolds(m_flow_pipe: float, r_in: float, fluid: Fluid) -> float:
        # Hydraulic diameter
        dia_hydraulic = 2.0 * r_in
        # Fluid velocity
        vol_flow_rate = m_flow_pipe / fluid.rho
        area_cr_inner = PI * r_in**2
        velocity = vol_flow_rate / area_cr_inner
        # Reynolds number
        return fluid.rho * velocity * dia_hydraulic / fluid.mu



from dataclasses import asdict, dataclass, field

from pygfunction.boreholes import Borehole

from ghedesigner.constants import RAD_TO_DEG
from ghedesigner.enums import DesignGeomType, FlowConfigType, TimestepType
from ghedesigner.ghe.design.base import DesignBase, GeometricConstraints
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.search.rowwise import RowWiseModifiedBisectionSearch
from ghedesigner.media import Fluid, Grout, Soil


@dataclass
class GeometricConstraintsRowWise(GeometricConstraints):
    """
    Geometric constraints for rowwise design algorithm
    """

    perimeter_spacing_ratio: float | None
    min_spacing: float
    max_spacing: float
    spacing_step: float | None
    min_rotation: float
    max_rotation: float
    rotate_step: float
    property_boundary: list[list[float]]
    no_go_boundaries: list[list[list[float]]] | None
    type: DesignGeomType = field(default=DesignGeomType.ROWWISE, init=False, repr=False)

    def to_input(self) -> dict:
        return {
            **asdict(self, dict_factory=lambda d: {k: v for k, v in d if k != "type"}),
            "min_rotation": self.min_rotation * RAD_TO_DEG,
            "max_rotation": self.max_rotation * RAD_TO_DEG,
            "method": self.type.name,
        }


class DesignRowWise(DesignBase):
    def __init__(
        self,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        start_month: int,
        end_month: int,
        max_eft: float,
        min_eft: float,
        max_height: float,
        min_height: float,
        continue_if_design_unmet: bool,
        max_boreholes: int | None,
        geometric_constraints: GeometricConstraintsRowWise,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        load_years=None,
    ) -> None:
        super().__init__(
            v_flow,
            borehole,
            fluid,
            pipe,
            grout,
            soil,
            start_month,
            end_month,
            max_eft,
            min_eft,
            max_height,
            min_height,
            continue_if_design_unmet,
            max_boreholes,
            geometric_constraints,
            hourly_extraction_ground_loads,
            method,
            flow_type,
            load_years,
        )
        self.geometric_constraints = geometric_constraints

    def find_design(self, disp=False) -> RowWiseModifiedBisectionSearch:
        if disp:
            title = "Find row-wise..."
            print(title + "\n" + len(title) * "=")
        return RowWiseModifiedBisectionSearch(
            self.v_flow,
            self.borehole,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.max_boreholes,
            self.min_height,
            self.max_height,
            self.continue_if_design_unmet,
            self.start_month,
            self.end_month,
            self.min_EFT_allowable,
            self.max_EFT_allowable,
            self.hourly_extraction_ground_loads,
            self.geometric_constraints,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="row-wise",
            load_years=self.load_years,
        )



from ghedesigner.building import Building
from ghedesigner.ghe.ground_heat_exchangers import GHE


class System:  # TODO: This isn't being used except by a unit test, should we keep it?
    def __init__(self) -> None:
        self.num_months = 0
        self.initial_loop_temp = 20.0
        self.building: Building | None = None
        self.ghe: GHE | None = None

    def set_simulation_parameters(self, num_months: int, initial_loop_temp: float = 20):
        self.num_months = num_months
        self.initial_loop_temp = initial_loop_temp

    def set_building(self, building: Building):
        self.building = building

    def set_ghe(self, ghe: GHE):
        self.ghe = ghe

    def simulate(self):
        flow_rate = 3.14
        ghe_exit_temp = self.initial_loop_temp
        num_hours = self.num_months * 30 * 24
        for hour_index in range(num_hours):
            building_outlet_temp = self.building.calculate(hour_index, ghe_exit_temp, flow_rate)
            ghe_exit_temp = self.ghe.calculate(hour_index, building_outlet_temp, flow_rate)



from copy import deepcopy
from typing import cast

from bhr.borehole import Borehole as BHRBorehole
from numpy import log, sqrt
from pygfunction.boreholes import Borehole

from ghedesigner.constants import PI, TWO_PI
from ghedesigner.enums import DoubleUTubeConnType
from ghedesigner.ghe.boreholes.base import GHEDesignerBoreholeBase
from ghedesigner.ghe.boreholes.single_u_borehole import SingleUTube
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil
from ghedesigner.utilities import solve_root


class GHEDesignerBoreholeWithMultiplePipes(GHEDesignerBoreholeBase):
    @staticmethod
    def calc_mass_flow_pipe(m_flow_borehole: float, config: DoubleUTubeConnType | None = None) -> float:
        if config == DoubleUTubeConnType.SERIES or config is None:
            return m_flow_borehole
        elif config == DoubleUTubeConnType.PARALLEL:
            return m_flow_borehole / 2.0
        else:
            raise ValueError(f"Invalid flow configuration: {config!s}")

    def equivalent_single_u_tube(
        self, vol_fluid: float, vol_pipe: float, resist_conv: float, resist_pipe: float, rho_cp_p: float
    ) -> SingleUTube:
        # Note: BHE can be double U-tube or coaxial heat exchanger

        # Compute equivalent single U-tube geometry
        n = 2
        r_p_i_prime = sqrt(vol_fluid / (n * PI))
        r_p_o_prime = sqrt((vol_fluid + vol_pipe) / (n * PI))
        k_p_prime = log(r_p_o_prime / r_p_i_prime) / (TWO_PI * n * resist_pipe)

        # Place single u-tubes at a B-spacing
        # Total horizontal space (m)
        # TODO: investigate why this deepcopy is required
        borehole = deepcopy(self.borehole)
        spacing = borehole.r_b * 2 - (n * r_p_o_prime * 2)
        # If the spacing is negative, then the borehole is not large enough,
        # therefore, the borehole will be increased if necessary
        if spacing <= 0.0:
            borehole.r_b -= spacing  # Add on the necessary spacing to fit
            spacing = (borehole.r_b * 2.0) / 10.0  # make spacing 1/10th of diameter
            borehole.r_b += spacing
        s = spacing / 3  # outer tube-to-tube shank spacing (m)

        # New pipe geometry
        roughness = self.pipe.roughness
        pipe = Pipe.init_single_u_tube(k_p_prime, rho_cp_p, r_p_i_prime * 2, r_p_o_prime * 2, s, roughness, 1)

        # Don't tie together the original and equivalent BHEs
        m_flow_borehole = self.m_flow_borehole
        fluid = self.fluid

        grout = Grout(self.grout.k, self.grout.rho_cp)
        soil = self.soil

        # Maintain the same mass flow rate so that the Rb/Rb* is not diverged from
        eq_single_u_tube = SingleUTube(m_flow_borehole, fluid, borehole, pipe, grout, soil)

        # The thermal conductivity of the pipe must now be varied such that R_fp is
        # equivalent to R_fp_prime
        def objective_pipe_conductivity(pipe_k: float):
            eq_single_u_tube.pipe.k = pipe_k
            eq_single_u_tube.calc_fluid_pipe_resistance()
            return eq_single_u_tube.R_fp - (resist_conv + resist_pipe)

        # Use Brent Quadratic to find the root
        # Define a lower and upper for pipe thermal conductivities
        k_p_lower = eq_single_u_tube.pipe.k / 100.0
        k_p_upper = eq_single_u_tube.pipe.k * 10.0

        # Solve for the pipe conductivity to make the resistance equal
        solve_root(
            eq_single_u_tube.pipe.k,
            objective_pipe_conductivity,
            lower=k_p_lower,
            upper=k_p_upper,
        )

        return eq_single_u_tube

    def match_effective_borehole_resistance(self, preliminary_new_single_u_tube: SingleUTube) -> SingleUTube:
        # Find the thermal conductivity that makes the borehole resistances equal

        # Define objective function for varying the grout thermal conductivity
        def objective_resistance(k_g_in: float):
            # update new tubes grout thermal conductivity and relevant parameters
            preliminary_new_single_u_tube.grout.k = k_g_in
            # Update Delta-circuit thermal resistances
            # Initialize stored_coefficients
            resist_bh_prime = preliminary_new_single_u_tube.calc_effective_borehole_resistance()
            resist_bh = self.calc_effective_borehole_resistance()
            return resist_bh - resist_bh_prime

        # Use Brent Quadratic to find the root
        # Define a lower and upper for thermal conductivities
        kg_lower = 1e-02
        kg_upper = 7.0
        k_g = solve_root(preliminary_new_single_u_tube.grout.k, objective_resistance, lower=kg_lower, upper=kg_upper)
        # Ensure the grout thermal conductivity is updated
        preliminary_new_single_u_tube.grout.k = k_g

        return preliminary_new_single_u_tube


class MultipleUTube(GHEDesignerBoreholeWithMultiplePipes):
    def __init__(
        self,
        m_flow_borehole: float,
        fluid: Fluid,
        borehole: Borehole,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        config=DoubleUTubeConnType.PARALLEL,
    ) -> None:
        # Ensure r_in and r_out are floats rather than list[float]
        if not isinstance(pipe.r_out, float) or not isinstance(pipe.r_in, float):
            raise TypeError("pipe r_in and r_out must be floats")

        super().__init__(m_flow_borehole, fluid, borehole, pipe, grout, soil)

        self.R_fp = 0.0
        self.fluid = fluid
        self.m_flow_borehole = m_flow_borehole
        self.m_flow_pipe = self.calc_mass_flow_pipe(self.m_flow_borehole, config)
        self.borehole = borehole
        self.pipe = pipe
        self.soil = soil
        self.grout = grout
        self.flow_config = config
        self.bhr_borehole = BHRBorehole()
        self.n_pipes = 2

        self.bhr_borehole.init_double_u_borehole(
            borehole_diameter=borehole.r_b * 2,
            pipe_outer_diameter=(2.0 * pipe.r_out),
            pipe_dimension_ratio=(2.0 * pipe.r_out) / (pipe.r_out - pipe.r_in),
            length=borehole.H,
            shank_space=(pipe.s / 2.0 + pipe.r_out),
            pipe_conductivity=pipe.k,
            grout_conductivity=grout.k,
            soil_conductivity=soil.k,
            fluid_type=self.fluid.name,
            fluid_concentration=self.fluid.concentration_percent / 100,
            boundary_condition="UNIFORM_BOREHOLE_WALL_TEMP",
            pipe_inlet_arrangement="ADJACENT",
        )

        self.calc_fluid_pipe_resistance()

    def calc_fluid_pipe_resistance(self) -> float:
        self.R_fp = self.bhr_borehole.calc_fluid_pipe_resist(self.m_flow_borehole, self.soil.ugt)
        return self.R_fp

    def calc_effective_borehole_resistance(self) -> float:
        resist_bh_effective = self.bhr_borehole.calc_bh_resist(self.m_flow_borehole, self.soil.ugt)
        return resist_bh_effective

    def u_tube_volumes(self) -> tuple[float, float]:
        # Compute volumes for U-tube geometry
        # Effective parameters
        n = self.n_pipes * 2  # Total number of tubes

        # Volumes
        vol_fluid = n * PI * (cast(float, self.pipe.r_in) ** 2)
        vol_pipe = n * PI * (cast(float, self.pipe.r_out) ** 2) - vol_fluid

        return vol_fluid, vol_pipe

    def to_single(self) -> SingleUTube:
        # Find an equivalent single U-tube given multiple U-tube geometry

        # Get effective parameters for the multiple u-tube
        vol_fluid, vol_pipe = self.u_tube_volumes()

        # TODO: check that the usage here for converting to an equivalent single u-tube is accurate
        resist_pipe = self.bhr_borehole.calc_pipe_cond_resist()
        resist_conv = self.R_fp - resist_pipe

        single_u_tube = self.equivalent_single_u_tube(vol_fluid, vol_pipe, resist_conv, resist_pipe, self.pipe.rho_cp)

        # Vary grout thermal conductivity to match effective borehole thermal resistance
        self.match_effective_borehole_resistance(single_u_tube)

        return single_u_tube



from math import cos, pi, sin
from typing import cast

from ghedesigner.enums import PipeType
from ghedesigner.media import ThermalProperty
from ghedesigner.utilities import check_arg_bounds

FloatOrFloats = float | list[float]
TupleOrTuples = tuple | list[tuple]


class Pipe(ThermalProperty):
    def __init__(self, pipe_type: PipeType, conductivity: float | tuple[float, float], rho_cp: float) -> None:
        super().__init__(conductivity, rho_cp)
        self.type = pipe_type

    @classmethod
    def init_from_dict(cls, pipe_type: PipeType, pipe_props: dict) -> "Pipe":
        rho_cp = pipe_props["rho_cp"]

        match pipe_type:
            case PipeType.SINGLEUTUBE:
                k = pipe_props["conductivity"]
                return Pipe.init_single_u_tube(
                    conductivity=k,
                    rho_cp=rho_cp,
                    inner_diameter=pipe_props["inner_diameter"],
                    outer_diameter=pipe_props["outer_diameter"],
                    shank_spacing=pipe_props["shank_spacing"],
                    roughness=pipe_props["roughness"],
                    num_pipes=pipe_props.get("num_pipes", 1),
                )
            case PipeType.DOUBLEUTUBESERIES:
                k = pipe_props["conductivity"]
                return Pipe.init_double_u_tube_series(
                    conductivity=k,
                    rho_cp=rho_cp,
                    inner_diameter=pipe_props["inner_diameter"],
                    outer_diameter=pipe_props["outer_diameter"],
                    shank_spacing=pipe_props["shank_spacing"],
                    roughness=pipe_props["roughness"],
                )
            case PipeType.DOUBLEUTUBEPARALLEL:
                k = pipe_props["conductivity"]
                return Pipe.init_double_u_tube_parallel(
                    conductivity=k,
                    rho_cp=rho_cp,
                    inner_diameter=pipe_props["inner_diameter"],
                    outer_diameter=pipe_props["outer_diameter"],
                    shank_spacing=pipe_props["shank_spacing"],
                    roughness=pipe_props["roughness"],
                )
            case PipeType.COAXIAL:
                k = pipe_props["conductivity_inner"], pipe_props["conductivity_outer"]
                return Pipe.init_coaxial(
                    conductivity=k,
                    rho_cp=rho_cp,
                    inner_pipe_d_in=pipe_props["inner_pipe_d_in"],
                    inner_pipe_d_out=pipe_props["inner_pipe_d_out"],
                    outer_pipe_d_in=pipe_props["outer_pipe_d_in"],
                    outer_pipe_d_out=pipe_props["outer_pipe_d_out"],
                    roughness=pipe_props["roughness"],
                )
            case _:
                raise ValueError(f"Pipe type {pipe_type} not recognized.")

    @classmethod
    def init_single_u_tube(
        cls,
        conductivity: float,
        rho_cp: float,
        inner_diameter: float,
        outer_diameter: float,
        shank_spacing: float,
        roughness: float,
        num_pipes: int = 1,
    ) -> "Pipe":
        check_arg_bounds(inner_diameter, outer_diameter, "inner_diameter", "outer_diameter")
        p = cls(PipeType.SINGLEUTUBE, conductivity, rho_cp)  # TODO: Untangle this a little more
        r_in = inner_diameter / 2.0
        r_out = outer_diameter / 2.0
        pipe_positions = Pipe.place_pipes(shank_spacing, r_out, num_pipes)
        p._finalize(pipe_positions, r_in, r_out, shank_spacing, roughness)
        return p

    @classmethod
    def init_double_u_tube_series(
        cls,
        conductivity: float,
        rho_cp: float,
        inner_diameter: float,
        outer_diameter: float,
        shank_spacing: float,
        roughness: float,
    ) -> "Pipe":
        check_arg_bounds(inner_diameter, outer_diameter, "inner_diameter", "outer_diameter")
        p = cls(PipeType.DOUBLEUTUBESERIES, conductivity, rho_cp)
        r_in = inner_diameter / 2.0
        r_out = outer_diameter / 2.0
        pipe_positions = Pipe.place_pipes(shank_spacing, r_out, 2)
        p._finalize(pipe_positions, r_in, r_out, shank_spacing, roughness)
        return p

    @classmethod
    def init_double_u_tube_parallel(
        cls,
        conductivity: float,
        rho_cp: float,
        inner_diameter: float,
        outer_diameter: float,
        shank_spacing: float,
        roughness: float,
    ) -> "Pipe":
        check_arg_bounds(inner_diameter, outer_diameter, "inner_diameter", "outer_diameter")
        p = cls(PipeType.DOUBLEUTUBEPARALLEL, conductivity, rho_cp)
        r_in = inner_diameter / 2.0
        r_out = outer_diameter / 2.0
        pipe_positions = Pipe.place_pipes(shank_spacing, r_out, 2)
        p._finalize(pipe_positions, r_in, r_out, shank_spacing, roughness)
        return p

    @classmethod
    def init_coaxial(
        cls,
        conductivity: tuple[float, float],
        rho_cp: float,
        inner_pipe_d_in: float,
        inner_pipe_d_out: float,
        outer_pipe_d_in: float,
        outer_pipe_d_out: float,
        roughness: float,
    ) -> "Pipe":
        check_arg_bounds(inner_pipe_d_in, inner_pipe_d_out, "inner_pipe_d_in", "inner_pipe_d_out")
        check_arg_bounds(outer_pipe_d_in, outer_pipe_d_out, "outer_pipe_d_in", "outer_pipe_d_out")
        p = cls(PipeType.COAXIAL, conductivity, rho_cp)
        # Note: This convention is different from pygfunction
        r_inner = [inner_pipe_d_in / 2.0, inner_pipe_d_out / 2.0]  # The radii of the inner pipe from in to out
        r_outer = [outer_pipe_d_in / 2.0, outer_pipe_d_out / 2.0]  # The radii of the outer pipe from in to out
        p._finalize((0, 0), r_inner, r_outer, 0, roughness)
        return p

    def _finalize(
        self,
        pos: TupleOrTuples,
        r_in: FloatOrFloats,
        r_out: FloatOrFloats,
        s: float,
        roughness: float,
    ) -> None:
        self.pos = pos  # Pipe positions either a list of tuples or tuple
        self.r_in = r_in  # Pipe inner radius (m) can be a float or list
        self.r_out = r_out  # Pipe outer radius (m) can be a float or list
        self.s = s  # Center pipe to center pipe shank spacing
        self.roughness = roughness  # Pipe roughness (m)
        self.n_pipes = int(len(pos) / 2) if isinstance(pos, list) else 1

    def as_dict(self) -> dict:
        output = {
            "base": super().as_dict(),
            "pipe_center_positions": str(self.pos),
            "shank_spacing_pipe_to_pipe": {"value": self.s, "units": "m"},
            "pipe_roughness": {"value": self.roughness, "units": "m"},
            "number_of_pipes": self.n_pipes,
        }
        if isinstance(self.r_in, float):
            r_in = cast(float, self.r_in)
            r_out = cast(float, self.r_out)
            output["pipe_inner_diameter"] = str(r_in * 2.0)
            output["pipe_outer_diameter"] = str(r_out * 2.0)
        else:
            r_in_list = cast(list[float], self.r_in)
            r_out_list = cast(list[float], self.r_out)
            output["pipe_inner_diameters"] = str([x * 2.0 for x in r_in_list])
            output["pipe_outer_diameters"] = str([x * 2.0 for x in r_out_list])
        return output

    @staticmethod
    def place_pipes(s, r_out, n_pipes):
        """Positions pipes in an axis-symmetric configuration."""
        shank_space = s / 2 + r_out
        dt = pi / float(n_pipes)
        pos = [(0.0, 0.0) for _ in range(2 * n_pipes)]
        for i in range(n_pipes):
            pos[2 * i] = (shank_space * cos(2.0 * i * dt + pi), shank_space * sin(2.0 * i * dt + pi))
            pos[2 * i + 1] = (shank_space * cos(2.0 * i * dt + pi + dt), shank_space * sin(2.0 * i * dt + pi + dt))
        return pos



from datetime import datetime
from math import floor
from typing import Any

from ghedesigner.enums import TimestepType
from ghedesigner.ghe.boreholes.base import GHEDesignerBoreholeBase
from ghedesigner.ghe.boreholes.coaxial_borehole import CoaxialPipe
from ghedesigner.ghe.ground_heat_exchangers import GHE
from ghedesigner.output.converters import hours_to_month


class JsonSerializer:
    """Generate the JSON-serializable SimulationSummary object."""

    @staticmethod
    def summary_object(
        ghe: GHE,
        search_tracker: Any,
        time_sec: float,
        project_name: str,
        notes: str,
        author: str,
        model_name: str,
        load_method: TimestepType,
    ) -> dict[str, Any]:
        def unit(v: float, u: str) -> dict[str, float | str]:
            return {"value": v, "units": u}

        # gFunction LTS
        g_lts = ghe.gFunction.g_lts
        titles = ["ln(t/ts)"] + [f"H: {h:0.2f} m" for h in g_lts] + [f"H: {ghe.bhe.borehole.H:0.2f} m"]
        data = []
        ghe_gf = ghe.gFunction.g_function_interpolation(ghe.b_spacing / ghe.bhe.borehole.H)[0]
        for i, t in enumerate(ghe.gFunction.log_time):
            row = [t] + [g_lts[h][i] for h in g_lts] + [ghe_gf[i]]
            data.append(row)

        # pipe geometry + reynolds
        if isinstance(ghe.bhe.pipe.r_out, float):
            pipe_geometry = {
                "pipe_outer_diameter": unit(ghe.bhe.pipe.r_out * 2.0, "m"),
                "pipe_inner_diameter": unit(ghe.bhe.pipe.r_in * 2.0, "m"),
            }
            reynolds = GHEDesignerBoreholeBase.compute_reynolds(
                ghe.bhe.m_flow_borehole, ghe.bhe.pipe.r_in, ghe.bhe.fluid
            )
        else:
            pipe_geometry = {
                "inner_pipe_inner_diameter": unit(ghe.bhe.pipe.r_in[0] * 2.0, "m"),
                "inner_pipe_outer_diameter": unit(ghe.bhe.pipe.r_in[1] * 2.0, "m"),
                "outer_pipe_inner_diameter": unit(ghe.bhe.pipe.r_out[0] * 2.0, "m"),
                "outer_pipe_outer_diameter": unit(ghe.bhe.pipe.r_out[1] * 2.0, "m"),
            }
            reynolds = CoaxialPipe.compute_reynolds_concentric(
                ghe.bhe.m_flow_borehole,
                ghe.bhe.r_in_out,
                ghe.bhe.r_out_in,
                ghe.bhe.fluid,
            )

        obj: dict[str, Any] = {
            "project_name": project_name,
            "notes": notes,
            "model_name": model_name,
            "simulation_time_stamp": datetime.now().strftime("%m/%d/%Y %H:%M:%S %p"),
            "simulation_author": author,
            "simulation_runtime": unit(time_sec, "s"),
            "design_selection_search_log": {
                "titles": ["Field", "Excess Temperature", "Max Temperature", "Min Temperature"],
                "units": [" ", "(C)", "(C)", "(C)"],
                "data": search_tracker,
            },
            "ghe_system": {
                "search_log": {"titles": titles, "units": None, "data": data},
                "active_borehole_length": unit(ghe.bhe.borehole.H, "m"),
                "borehole_diameter": unit(ghe.bhe.borehole.r_b * 2.0, "m"),
                "borehole_buried_depth": unit(ghe.bhe.borehole.D, "m"),
                "borehole_spacing": unit(ghe.b_spacing, "m"),
                "total_drilling": unit(ghe.bhe.borehole.H * len(ghe.gFunction.bore_locations), "m"),
                "field_type": ghe.field_type,
                "field_specifier": ghe.fieldSpecifier,
                "number_of_boreholes": len(ghe.gFunction.bore_locations),
                "shank_spacing": unit(ghe.bhe.pipe.s, "m"),
                "pipe_geometry": pipe_geometry,
                "pipe_roughness": unit(ghe.bhe.pipe.roughness, "m"),
                "pipe_thermal_conductivity": unit(ghe.bhe.pipe.k, "W/m-K"),
                "pipe_volumetric_heat_capacity": unit(ghe.bhe.pipe.rho_cp / 1000, "kJ/m3-K"),
                "grout_thermal_conductivity": unit(ghe.bhe.grout.k, "W/m-K"),
                "grout_volumetric_heat_capacity": unit(ghe.bhe.grout.rho_cp / 1000, "kJ/m3-K"),
                "reynolds_number": reynolds,
                "effective_borehole_resistance": unit(ghe.bhe.calc_effective_borehole_resistance(), "W/m-K"),
                # TODO: are the units right here?
                "soil_thermal_conductivity": unit(ghe.bhe.soil.k, "W/m-K"),
                "soil_volumetric_heat_capacity": unit(ghe.bhe.soil.rho_cp / 1000, "kJ/m3-K"),
                "soil_undisturbed_ground_temp": unit(ghe.bhe.soil.ugt, "C"),
                "fluid_volumetric_heat_capacity": unit(ghe.bhe.fluid.rho_cp / 1000, "kJ/m3-K"),
                "fluid_thermal_conductivity": unit(ghe.bhe.fluid.k, "W/m-K"),
                "fluid_viscosity": unit(ghe.bhe.fluid.mu, "Pa-s"),
                "fluid_mixture": ghe.bhe.fluid.name,
                "fluid_density": unit(ghe.bhe.fluid.rho, "kg/m3"),
                "fluid_mass_flow_rate_per_borehole": unit(ghe.bhe.m_flow_borehole, "kg/s"),
            },
            "simulation_parameters": {
                "simulation_load_method": load_method.name,
            },
            "simulation_results": {},
        }

        # Optional convection coefficient
        if hasattr(ghe.bhe, "h_f"):
            obj["ghe_system"]["fluid_convection_coefficient"] = unit(ghe.bhe.h_f, "W/m-K")

        # Monthly loads (when hybrid loads are provided)
        if getattr(ghe, "hybrid_load", None):
            monthly_load_values: list[list[Any]] = []
            n_months = len(ghe.hybrid_load.monthly_cl) - 1
            n_years = int(n_months / 12)
            months = n_years * [
                "January",
                "February",
                "March",
                "April",
                "May",
                "June",
                "July",
                "August",
                "September",
                "October",
                "November",
                "December",
            ]
            for i in range(1, n_months + 1):
                monthly_load_values.append(
                    [
                        months[i - 1],
                        ghe.hybrid_load.monthly_hl[i],
                        ghe.hybrid_load.monthly_cl[i],
                        ghe.hybrid_load.monthly_peak_hl[i],
                        ghe.hybrid_load.monthly_peak_hl_duration[i],
                        ghe.hybrid_load.monthly_peak_cl[i],
                        ghe.hybrid_load.monthly_peak_cl_duration[i],
                    ]
                )
            obj["ghe_system"]["glhe_monthly_loads"] = {
                "titles": [
                    "Month",
                    "Total Heating",
                    "Total Cooling",
                    "Peak Heating",
                    "PH Duration",
                    "Peak Cooling",
                    "PC Duration",
                ],
                "units": ["", "kWh", "kWh", "kW", "hr", "kW", "hr"],
                "data": monthly_load_values,
            }

            # Peak EFTs + monthly temperature summary
            out_array = []
            last_month = -1
            month_tb_vals: list[float] = []
            month_eft_vals: list[float] = []
            for tv, d_tb, eft in zip(ghe.times, ghe.dTb, ghe.hp_eft):
                current_month = floor(hours_to_month(tv))
                if current_month == last_month:
                    month_tb_vals.append(d_tb)
                    month_eft_vals.append(eft)
                else:
                    if month_tb_vals:
                        previous_temp = ghe.bhe.soil.ugt
                        out_array.append(
                            [
                                current_month,
                                previous_temp + month_tb_vals[-1],
                                max(month_eft_vals),
                                min(month_eft_vals),
                            ]
                        )
                    last_month = current_month
                    month_tb_vals = [d_tb]
                    month_eft_vals = [eft]

            max_eft = max(ghe.hp_eft)
            min_eft = min(ghe.hp_eft)
            max_eft_time = hours_to_month(ghe.times[ghe.hp_eft.index(max_eft)])
            min_eft_time = hours_to_month(ghe.times[ghe.hp_eft.index(min_eft)])

            obj["simulation_results"] = {
                "max_hp_eft": unit(max_eft, "C"),
                "max_hp_eft_time": unit(max_eft_time, "months"),
                "min_hp_eft": unit(min_eft, "C"),
                "min_hp_eft_time": unit(min_eft_time, "months"),
                "monthly_temp_summary": {
                    "titles": ["Time", "BH Wall Temp", "Max HP EFT", "Min HP EFT"],
                    "units": ["(months)", "(C)", "(C)", "(C)"],
                    "data": out_array,
                },
            }

        return obj



import importlib.metadata
from math import pi, tau

VERSION = importlib.metadata.version("ghedesigner")

PI = pi
DEG_TO_RAD = pi / 180.0
RAD_TO_DEG = 180.0 / pi
PI_OVER_2 = pi / 2.0
TWO_PI = tau
HRS_IN_DAY = 24
SEC_IN_HR = 3600
MONTHS_IN_YEAR = 12



from math import atan, inf, pi, sin, sqrt

import numpy as np

from ghedesigner.constants import PI_OVER_2, TWO_PI


class Shapes:
    def __init__(self, c) -> None:
        """
        constructs a shape object
        """
        self.c = np.array(c)
        # print(c)
        self.max_x = np.max(self.c[:, 0])
        self.min_x = np.min(self.c[:, 0])
        self.max_y = np.max(self.c[:, 1])
        self.min_y = np.min(self.c[:, 1])
        self.area = self.get_area()
        self.arc_lengths = self._calc_arc_lengths()
        self.total_arc_length = sum(self.arc_lengths)
        self.piecewise_maxes = self._prep_arc_length_piecewise_function()

    def _calc_arc_lengths(self):
        lengths = []
        for i in range(len(self.c)):
            if i == len(self.c) - 1:
                c1 = self.c[i]
                c2 = self.c[0]
            else:
                c1 = self.c[i]
                c2 = self.c[i + 1]
            lengths.append(distance(c1, c2))
        return lengths

    def _prep_arc_length_piecewise_function(self):
        piecewise_maxes = [0.0]
        current_length = 0
        maximum_length = self.total_arc_length
        for arc_length in self.arc_lengths:
            current_length += arc_length
            piecewise_maxes.append(current_length / maximum_length)
        return piecewise_maxes

    def _get_point_along_arc(self, ratio, arc_index):
        if arc_index == len(self.c) - 1:
            c1 = self.c[arc_index]
            c2 = self.c[0]
        else:
            c1 = self.c[arc_index]
            c2 = self.c[arc_index + 1]
        return ratio * (c2[0] - c1[0]) + c1[0], ratio * (c2[1] - c1[1]) + c1[1]

    def get_point_along_curve(self, relative_distance_along_curve):
        larger_index = next(i for i, x in enumerate(self.piecewise_maxes) if x >= relative_distance_along_curve)
        arc_ratio = (relative_distance_along_curve - self.piecewise_maxes[larger_index - 1]) / (
            self.piecewise_maxes[larger_index] - self.piecewise_maxes[larger_index - 1]
        )
        return self._get_point_along_arc(arc_ratio, larger_index - 1)

    def line_intersect(self, xy, rotate=0, intersection_tolerance=1e-6):
        """
        returns the intersections between a line segment and the shape

        Parameters
        -----------
        :param xy: [float,float,float,float]
            the x,y values of both endpoints of the line segment
        :param rotate:
        :param intersection_tolerance:

        :return: [[float]]
            the x,y values of the intersections
        """
        x1, y1, x2, y2 = xy
        minx, maxx = (x1, x2) if x1 < x2 else (x2, x1)
        miny, maxy = (y1, y2) if y1 < y2 else (y2, y1)

        r_a = []

        # Check if they fall within the same rectangular area - note: calling functions seem to assume
        # that this is an infinite line intersection check (rather than line segment).
        # if maxx < self.min_x or minx > self.max_x or maxy < self.min_y or miny > self.max_y:
        # return r_a

        for i in range(len(self.c)):
            if i == len(self.c) - 1:
                c1 = self.c[len(self.c) - 1]
                c2 = self.c[0]
                r = vector_intersect(
                    [c1[0], c1[1], c2[0], c2[1]],
                    [x1, y1, x2, y2],
                    intersection_tolerance,
                )

                if len(r) == 1:
                    r = r[0]
                    if (
                        (r[0] - max(c2[0], c1[0])) > intersection_tolerance
                        or (r[0] - min(c2[0], c1[0])) < -1 * intersection_tolerance
                        or (r[1] - max(c2[1], c1[1])) > intersection_tolerance
                        or (r[1] - min(c2[1], c1[1])) < -1 * intersection_tolerance
                    ):
                        continue
                    r_a.append(r)
            else:
                c1 = self.c[i]
                c2 = self.c[i + 1]
                r = vector_intersect(
                    [c1[0], c1[1], c2[0], c2[1]],
                    [x1, y1, x2, y2],
                    intersection_tolerance,
                )
                if len(r) == 1:
                    r = r[0]
                    if (
                        (r[0] - max(c2[0], c1[0])) > intersection_tolerance
                        or (r[0] - min(c2[0], c1[0])) < -1 * intersection_tolerance
                        or (r[1] - max(c2[1], c1[1])) > intersection_tolerance
                        or (r[1] - min(c2[1], c1[1])) < -1 * intersection_tolerance
                    ):
                        continue
                    r_a.append(r)
        # print("x value: %f, r values:"%x1)
        # print(r_a)

        r_a = sort_intersections(r_a, rotate)
        # print(r_a)
        return r_a

    def point_intersect(self, xy):
        """
        returns whether the given point is inside the rectangle

        Parameters
        -----------
        :param xy: [float,float]
            x,y value of point

        :return: boolean
            true if inside, false if not
        """
        x, y = xy
        if (x > self.max_x or x < self.min_x) or (y > self.max_y or y < self.min_y):
            # print("Returning False b/c outside of box")
            return False
        far_x = self.min_x - 10
        inters = self.line_intersect([far_x, y, far_x + 1, y])
        # print(inters)
        inters = [inter for inter in inters if inter[0] <= x]
        # print("x: %f"%x,inters)
        if len(inters) == 1:
            # print("Returning True")
            return True
        i = 0
        while i < len(inters):
            for vert in self.c:
                if inters[i][0] == vert[0] and inters[i][1] == vert[1]:
                    inters.pop(i)
                    i -= 1
                    break
            i += 1
        return len(inters) % 2 != 0
        # False if even, True if odd

    def get_area(self):
        """
        returns area of shape

        :return: float
            area of shape
        """
        area_sum = 0
        for i in range(len(self.c)):
            if i == len(self.c) - 1:
                area_sum += self.c[len(self.c) - 1][0] * self.c[0][1] - (self.c[len(self.c) - 1][1] * self.c[0][0])
                continue
            area_sum += self.c[i][0] * self.c[i + 1][1] - (self.c[i][1] * self.c[i + 1][0])
        return 0.5 * area_sum


def sort_intersections(r_a, rotate):
    if len(r_a) == 0:
        return r_a
    vals = [0] * len(r_a)
    for i, inter in enumerate(r_a):
        phi = PI_OVER_2 if inter[0] == 0 else atan(inter[1] / inter[0])
        dist_inter = sqrt(inter[1] ** 2 + inter[0] ** 2)
        ref_ang = PI_OVER_2 - phi
        # sign = 1
        if phi > PI_OVER_2:
            if phi > pi:  # noqa: SIM108
                ref_ang = TWO_PI - phi if phi > 3 * PI_OVER_2 else 3.0 * PI_OVER_2 - phi
            else:
                ref_ang = pi - phi
        # if phi > pi/2 + rotate and phi < 3*pi/2 + rotate:
        # sign = -1
        vals[i] = dist_inter * sin(ref_ang + rotate)
    zipped = sorted(zip(vals, r_a))
    r_a = [row for _, row in zipped]
    return r_a


def vector_intersect(l1, l2, intersection_tolerance):
    """
     gives the intersection between two line segments

    Parameters
    -----------
    :param l1: [[float]]
        endpoints of first line segment
    :param l2: [[float]]
        endpoints of the second line segment
    :param intersection_tolerance:

    :return: [float,float]
        x,y values of intersection (returns None if there is none)
    """
    x11, y11, x12, y12 = l1
    x21, y21, x22, y22 = l2
    if x12 - x11 == 0:
        a1 = inf
    else:
        a1 = (y12 - y11) / (x12 - x11)
        c1 = y11 - x11 * a1
    if x22 - x21 == 0:
        a2 = inf
    else:
        a2 = (y22 - y21) / (x22 - x21)
        c2 = y21 - x21 * a2
    if inf in (a1, a2):
        if a1 == inf and a2 == inf:
            if abs(x11 - x21) < intersection_tolerance:
                return [[x11, y11], [x12, y12]]
            else:
                return []
        elif a1 == inf:
            return [[x11, a2 * x11 + c2]]
        else:
            return [[x21, a1 * x21 + c1]]
    if abs(a1 - a2) <= intersection_tolerance:
        if abs(y22 - (a1 * x22 + c1)) <= intersection_tolerance:
            # return [[x11,y11],[x12,y12]]
            return []
        else:
            return []
    rx = (c2 - c1) / (a1 - a2)
    ry = a1 * (c2 - c1) / (a1 - a2) + c1
    return [[rx, ry]]


def distance(pt_1, pt_2) -> float:
    return sqrt((pt_1[0] - pt_2[0]) ** 2 + (pt_1[1] - pt_2[1]) ** 2)


def point_polygon_check(contour, point, on_edge_tolerance=1e-6):
    """
    Mimics pointPolygonTest from OpenCV-Python

    Adapted from the methods outlined in the links below.
    https://stackoverflow.com/a/63436180/5965685
    https://stackoverflow.com/a/17693146/5965685

    :param contour: list of tuples containing (x, y) contour boundary points
    :param point: tuple containing the (x, y) point to test
    :param on_edge_tolerance: float value representing the distance tolerance
                               to determine whether the test point is on the edge of the polygon.
                               Points within this distance from the edge are considered "on the edge".
                               Defaults to 0.001.


    :returns: -1 if outside, 0 if on edge, 1 if inside
    :rtype: int
    """

    # check if on edge
    # use Pythagoras to check whether the distance between the test point
    # and the line vertices add to the distance between the line vertices
    # if they are within tolerance, the point is co-linear
    # if not, it is off the line

    for idx, vertex in enumerate(contour):
        v1 = contour[idx - 1]
        v2 = vertex
        test_dist = distance(v1, point) + distance(v2, point)
        v12_dist = distance(v1, v2)

        if abs(test_dist - v12_dist) < on_edge_tolerance:
            return 0

    # if made it to here, not on edge and check if inside/outside
    def between(p, a, b) -> bool:
        return ((p >= a) and (p <= b)) or ((p <= a) and (p >= b))

    inside = True
    px = point[0]
    py = point[1]

    for idx, vertex in enumerate(contour):
        v1 = contour[idx - 1]
        v2 = vertex
        v1x = v1[0]
        v1y = v1[1]
        v2x = v2[0]
        v2y = v2[1]

        if between(py, v1y, v2y):  # points inside vertical range
            if ((py == v1y) and (v2y >= v1y)) or ((py == v2y) and (v1y >= v2y)):
                continue

            # calc cross product `PA X PB`, P lies on left side of AB if c > 0
            c = (v1x - px) * (v2y - py) - (v2x - px) * (v1y - py)

            if c == 0:
                return 0

            if (v1y < v2y) == (c > 0):
                inside = not inside

    return -1 if inside else 1



from ghedesigner import (
    building,
    constants,
    enums,
    heat_pump,
    media,
    output,
    system,
    utilities,
    validate,
)
from ghedesigner.constants import VERSION
from ghedesigner.ghe import *



from math import inf, sqrt

import numpy as np
from pygfunction.boreholes import Borehole

from ghedesigner.enums import FlowConfigType, TimestepType
from ghedesigner.ghe.gfunction import calc_g_func_for_multiple_lengths
from ghedesigner.ghe.ground_heat_exchangers import GHE
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.rowwise import field_optimization_fr, field_optimization_wp_space_fr, gen_shape
from ghedesigner.media import Fluid, Grout, Soil
from ghedesigner.utilities import borehole_spacing, eskilson_log_times


# This is the search algorithm used for finding row-wise fields
class RowWiseModifiedBisectionSearch:
    def __init__(
        self,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        max_boreholes: int | None,  # TODO: Unused
        min_height: float,
        max_height: float,
        continue_if_design_unmet: bool,
        start_month: int,
        end_month: int,
        min_eft: float,
        max_eft: float,
        hourly_extraction_ground_loads: list,
        geometric_constraints,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        max_iter: int = 10,
        disp: bool = False,
        search: bool = True,
        advanced_tracking: bool = True,
        field_type: str = "rowwise",
        load_years=None,
    ) -> None:
        # Take the lowest part of the coordinates domain to be used for the
        # initial setup
        if load_years is None:
            load_years = [2019]
        self.load_years = load_years
        self.fluid = fluid
        self.pipe = pipe
        self.grout = grout
        self.soil = soil
        self.borehole = borehole
        self.geometricConstraints = geometric_constraints
        self.searchTracker: list[list] = []
        self.field_type = field_type
        # Flow rate tracking
        self.max_boreholes = max_boreholes
        self.max_height = max_height
        self.max_eft = max_eft
        self.min_eft = min_eft
        self.min_height = min_height
        self.max_height = max_height
        self.start_month = start_month
        self.end_month = end_month
        self.continue_if_design_unmet = continue_if_design_unmet
        self.v_flow = v_flow
        self.flow_type = flow_type
        self.method = method
        self.log_time = eskilson_log_times()
        self.bhe_type = pipe.type
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.max_iter = max_iter
        self.disp = disp
        self.ghe: GHE | None = None
        self.calculated_temperatures: dict = {}
        if advanced_tracking:
            self.advanced_tracking = [["TargetSpacing", "Field Specifier", "nbh", "ExcessTemperature"]]
            self.checkedFields: list[np.ndarray[tuple[np.float64, np.float64]]] = []
        if search:
            self.selected_coordinates, self.selected_specifier = self.search()
            self.initialize_ghe(self.selected_coordinates, self.max_height, field_specifier=self.selected_specifier)

    def retrieve_flow(self, coordinates, rho):
        if self.flow_type == FlowConfigType.BOREHOLE:
            v_flow_system = self.v_flow * len(coordinates)
            # Total fluid mass flow rate per borehole (kg/s)
            m_flow_borehole = self.v_flow / 1000.0 * rho
        elif self.flow_type == FlowConfigType.SYSTEM:
            v_flow_system = self.v_flow
            v_flow_borehole = self.v_flow / len(coordinates)
            m_flow_borehole = v_flow_borehole / 1000.0 * rho
        else:
            raise ValueError("The flow argument should be either `borehole` or `system`.")
        return v_flow_system, m_flow_borehole

    def initialize_ghe(self, coordinates, h, field_specifier="N/A"):
        v_flow_system, m_flow_borehole = self.retrieve_flow(coordinates, self.fluid.rho)

        self.borehole.H = h
        borehole = self.borehole
        fluid = self.fluid
        pipe = self.pipe
        grout = self.grout
        soil = self.soil

        b = borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = calc_g_func_for_multiple_lengths(
            b,
            [borehole.H],
            borehole.r_b,
            borehole.D,
            m_flow_borehole,
            self.bhe_type,
            self.log_time,
            coordinates,
            fluid,
            pipe,
            grout,
            soil,
        )

        # Initialize the GHE object
        self.ghe = GHE(
            v_flow_system,
            b,
            self.bhe_type,
            fluid,
            borehole,
            pipe,
            grout,
            soil,
            g_function,
            self.start_month,
            self.end_month,
            self.hourly_extraction_ground_loads,
            field_type=self.field_type,
            field_specifier=field_specifier,
        )

    def calculate_excess(self, coordinates, h, design_max_eft, design_min_eft, field_specifier="N/A"):
        self.initialize_ghe(coordinates, h, field_specifier=field_specifier)
        # Simulate after computing just one g-function
        max_hp_eft, min_hp_eft = self.ghe.simulate(method=self.method)
        t_excess = self.ghe.cost(max_hp_eft, min_hp_eft, design_max_eft, design_min_eft)
        self.searchTracker.append([field_specifier, t_excess, max_hp_eft, min_hp_eft])

        return t_excess

    def search(self):
        spacing_start = self.geometricConstraints.min_spacing
        spacing_stop = self.geometricConstraints.max_spacing
        spacing_step = self.geometricConstraints.spacing_step
        rotate_step = self.geometricConstraints.rotate_step
        prop_bound, ng_zones = gen_shape(
            self.geometricConstraints.property_boundary, self.geometricConstraints.no_go_boundaries
        )
        rotate_start = self.geometricConstraints.min_rotation
        rotate_stop = self.geometricConstraints.max_rotation
        perimeter_spacing_ratio = self.geometricConstraints.perimeter_spacing_ratio

        use_perimeter = not (perimeter_spacing_ratio is None or perimeter_spacing_ratio == 0)
        skip_secondary_interval_search = spacing_step is None or spacing_step == 0

        selected_coordinates = None
        selected_specifier = None
        selected_temp_excess = None
        selected_spacing = None

        # Check The Upper and Lower Bounds

        # Generate Fields
        if use_perimeter:
            upper_field, upper_field_specifier = field_optimization_wp_space_fr(
                perimeter_spacing_ratio,
                spacing_start,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )
            lower_field, lower_field_specifier = field_optimization_wp_space_fr(
                perimeter_spacing_ratio,
                spacing_stop,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )
        else:
            upper_field, upper_field_specifier = field_optimization_fr(
                spacing_start,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )
            lower_field, lower_field_specifier = field_optimization_fr(
                spacing_stop,
                rotate_step,
                prop_bound,
                ng_zones=ng_zones,
                rotate_start=rotate_start,
                rotate_stop=rotate_stop,
            )

        # Get Excess Temperatures
        t_upper = self.calculate_excess(
            upper_field, self.max_height, self.max_eft, self.min_eft, field_specifier=upper_field_specifier
        )
        t_lower = self.calculate_excess(
            lower_field, self.max_height, self.max_eft, self.min_eft, field_specifier=lower_field_specifier
        )

        if self.advanced_tracking:
            self.advanced_tracking.append([spacing_start, upper_field_specifier, len(upper_field), t_upper])
            self.advanced_tracking.append([spacing_stop, lower_field_specifier, len(lower_field), t_lower])
            self.checkedFields.append(upper_field)
            self.checkedFields.append(lower_field)

        # If the excess temperature is >0 utilizing the largest field and largest depth, then notify the user that
        # the given constraints cannot find a satisfactory field.
        if t_upper > 0.0 and t_lower > 0.0:
            condition_msg = (
                "The optimal design requires more or deeper boreholes \n"
                "than what is possible based on the current design parameters. \n"
                "Consider increasing the available land area, \n"
                "increasing the maximum borehole depth, \n"
                "or decreasing the maximum borehole spacing."
            )
            print(condition_msg)
            if self.continue_if_design_unmet:
                print("Largest available configuration selected.")
                return upper_field, upper_field_specifier
            else:
                raise ValueError("Search failed.")

        # If the excess temperature is > 0 when utilizing the largest field and depth but < 0 when using the largest
        # depth and smallest field, then fields should be searched between the two target depths.
        elif t_upper < 0.0 < t_lower:
            # This search currently works by doing a slightly modified bisection search where the "steps" are the set
            # by the "spacing_step" variable. The steps are used to check fields on either side of the field found by
            # doing a normal bisection search. These extra fields are meant to help prevent falling into local minima
            # (although this will still happen sometimes).
            i = 0
            spacing_high = spacing_start
            spacing_low = spacing_stop
            low_e = t_upper
            high_e = t_lower
            spacing_m = (spacing_stop + spacing_start) * 0.5
            while i < self.max_iter:
                print("Bisection Search Iteration: ", i)
                # Getting Three Middle Field
                if use_perimeter:
                    f1, f1_specifier = field_optimization_wp_space_fr(
                        perimeter_spacing_ratio,
                        spacing_m,
                        rotate_step,
                        prop_bound,
                        ng_zones=ng_zones,
                        rotate_start=rotate_start,
                        rotate_stop=rotate_stop,
                    )
                else:
                    f1, f1_specifier = field_optimization_fr(
                        spacing_m,
                        rotate_step,
                        prop_bound,
                        ng_zones=ng_zones,
                        rotate_start=rotate_start,
                        rotate_stop=rotate_stop,
                    )

                # Getting the three field's excess temperature
                t_e1 = self.calculate_excess(
                    f1, self.max_height, self.max_eft, self.min_eft, field_specifier=f1_specifier
                )

                if self.advanced_tracking:
                    self.advanced_tracking.append([spacing_m, f1_specifier, len(f1), t_e1])
                    self.checkedFields.append(f1)
                if t_e1 <= 0.0:
                    spacing_high = spacing_m
                    high_e = t_e1
                    selected_specifier = f1_specifier
                    selected_coordinates = f1
                else:
                    spacing_low = spacing_m
                    low_e = t_e1

                spacing_m = (spacing_low + spacing_high) * 0.5
                error_tolerance = 1e-10
                if abs(low_e - high_e) < error_tolerance:
                    break

                i += 1

            if skip_secondary_interval_search:
                return selected_coordinates, selected_specifier
            else:
                # Now Check fields that have a higher target spacing to double-check that none of them would work:
                spacing_l = spacing_step + spacing_high
                target_spacings = []
                current_spacing = spacing_high

                # TODO: this was an argument, but was never used.
                exhaustive_fields_to_check = 10
                spacing_change = (spacing_l - current_spacing) / exhaustive_fields_to_check
                while current_spacing <= spacing_l:
                    target_spacings.append(current_spacing)
                    current_spacing += spacing_change
                best_field = None
                best_drilling = inf
                best_excess = None
                best_spacing = None
                for ts in target_spacings:
                    if use_perimeter:
                        field, f_s = field_optimization_wp_space_fr(
                            perimeter_spacing_ratio,
                            ts,
                            rotate_step,
                            prop_bound,
                            ng_zones=ng_zones,
                            rotate_start=rotate_start,
                            rotate_stop=rotate_stop,
                        )
                    else:
                        field, f_s = field_optimization_fr(
                            ts,
                            rotate_step,
                            prop_bound,
                            ng_zones=ng_zones,
                            rotate_start=rotate_start,
                            rotate_stop=rotate_stop,
                        )

                    t_e = self.calculate_excess(field, self.max_height, self.max_eft, self.min_eft, field_specifier=f_s)

                    if self.advanced_tracking:
                        self.advanced_tracking.append([ts, f_s, len(field), t_e])
                        self.checkedFields.append(field)

                    self.initialize_ghe(field, self.max_height, field_specifier=f_s)
                    self.ghe.compute_g_functions(self.min_height, self.max_height)
                    self.ghe.size(
                        method=TimestepType.HYBRID,
                        max_height=self.max_height,
                        min_height=self.min_height,
                        design_max_eft=self.max_eft,
                        design_min_eft=self.min_eft,
                    )
                    total_drilling = self.ghe.bhe.borehole.H * len(field)

                    if best_field is None:
                        best_field = field
                        best_drilling = total_drilling
                        best_excess = t_e
                        best_spacing = ts
                    elif t_e <= 0.0 and total_drilling < best_drilling:
                        best_drilling = total_drilling
                        best_field = field
                        best_excess = t_e
                        best_spacing = ts
                selected_coordinates = best_field
                selected_temp_excess = best_excess
                selected_spacing = best_spacing

        # If the excess temperature is < 0 when utilizing the largest depth and the smallest field, it is most likely
        # in the user's best interest to return a field smaller than the smallest one. This is done by removing
        # boreholes from the field.
        elif t_lower < 0.0 and t_upper < 0.0:
            original_coordinates = lower_field

            # Function For Sorting Boreholes Based on Proximity to a Point
            def point_sort(target_point, other_points, method="ascending"):
                def dist(o_p):
                    return sqrt(
                        (target_point[0] - o_p[0]) * (target_point[0] - o_p[0])
                        + (target_point[1] - o_p[1]) * (target_point[1] - o_p[1])
                    )

                distances = map(dist, other_points)
                if method == "ascending":
                    return [x for _, x in sorted(zip(distances, other_points))]
                elif method == "descending":
                    return [x for _, x in sorted(zip(distances, other_points), reverse=True)]

            # TODO: b_r_removal_method was an argument but it was never used
            # if b_r_removal_method == "CloseToCorner":
            starting_field = point_sort(original_coordinates[0], lower_field, method="descending")
            # elif b_r_removal_method == "CloseToPoint":
            #     starting_field = point_sort([0.0, 0.0], lower_field, method="descending")
            # elif b_r_removal_method == "FarFromPoint":
            #     starting_field = point_sort([0.0, 0.0], lower_field, method="ascending")
            # elif b_r_removal_method == "RowRemoval":
            #     starting_field = lower_field
            # else:
            #     msg = b_r_removal_method + " is not a valid method for removing boreholes."
            #     msg += "The valid methods are: CloseToCorner, CloseToPoint, FarFromPoint, and RowRemoval."
            #     raise ValueError(msg)

            # Check if a 1X1 field is satisfactory
            t_e_single = self.calculate_excess(
                [[0, 0]], self.max_height, self.max_eft, self.min_eft, field_specifier="1X1"
            )

            if self.advanced_tracking:
                self.advanced_tracking.append(["N/A", "1X1", 1, t_e_single])
                self.checkedFields.append([[0, 0]])
            if t_e_single <= 0:
                selected_temp_excess = t_e_single
                selected_specifier = "1X1"
                selected_coordinates = starting_field[len(starting_field) - 1 :]
                selected_spacing = spacing_stop
            else:
                # Perform a bisection search between nbh values to find the smallest satisfactory field
                nbh_max = len(starting_field)
                nbh_min = 1
                nbh_start = nbh_max
                # continueLoop = True
                # highT_e = T_lower
                selected_specifier = lower_field_specifier
                i = 0
                while i < self.max_iter:
                    nbh = (nbh_max + nbh_min) // 2
                    current_field = starting_field[nbh_start - nbh :]
                    f_s = lower_field_specifier + f"_BR{nbh_start - nbh}"
                    t_e = self.calculate_excess(
                        current_field, self.max_height, self.max_eft, self.min_eft, field_specifier=f_s
                    )
                    if self.advanced_tracking:
                        self.advanced_tracking.append([spacing_stop, lower_field_specifier + "_" + str(nbh), nbh, t_e])
                        self.checkedFields.append(current_field)
                    if t_e <= 0.0:
                        # highT_e = T_e
                        nbh_max = nbh
                        selected_coordinates = current_field
                        selected_specifier = f_s
                        selected_temp_excess = t_e
                        selected_spacing = spacing_stop
                    else:
                        nbh_min = nbh
                    if (nbh_max - nbh_min) <= 1:
                        break
                    i += 1
        # If none of the options above have been true, then there is most likely an issue with the excess temperature
        # calculation.
        else:
            msg = (
                "There seems to be an issue calculating excess temperatures. Check that you have the correct \n"
                "package version. If this is a recurring issue, please contact the current package management for \n"
                "assistance."
            )
            raise ValueError(msg)
        if self.advanced_tracking:
            self.advanced_tracking.append(
                [
                    selected_spacing,
                    selected_specifier,
                    len(selected_coordinates),
                    selected_temp_excess,
                ]
            )
            self.checkedFields.append(selected_coordinates)
        return selected_coordinates, selected_specifier



from dataclasses import asdict, dataclass, field
from math import floor

from pygfunction.boreholes import Borehole

from ghedesigner.enums import DesignGeomType, FlowConfigType, TimestepType
from ghedesigner.ghe.design.base import DesignBase, GeometricConstraints
from ghedesigner.ghe.domains import square_and_near_square
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.search.bisection_1d import Bisection1D
from ghedesigner.media import Fluid, Grout, Soil


@dataclass
class GeometricConstraintsNearSquare(GeometricConstraints):
    """
    Geometric constraints for near square design algorithm
    """

    b: float
    length: float
    type: DesignGeomType = field(default=DesignGeomType.NEARSQUARE, init=False, repr=False)

    def to_input(self) -> dict:
        return {
            **asdict(self, dict_factory=lambda d: {k: v for k, v in d if k != "type"}),
            "method": self.type.name,
        }


class DesignNearSquare(DesignBase):
    def __init__(
        self,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        start_month: int,
        end_month: int,
        max_eft: float,
        min_eft: float,
        max_height: float,
        min_height: float,
        continue_if_design_unmet: bool,
        max_boreholes: int | None,
        geometric_constraints: GeometricConstraintsNearSquare,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        load_years=None,
    ) -> None:
        super().__init__(
            v_flow,
            borehole,
            fluid,
            pipe,
            grout,
            soil,
            start_month,
            end_month,
            max_eft,
            min_eft,
            max_height,
            min_height,
            continue_if_design_unmet,
            max_boreholes,
            geometric_constraints,
            hourly_extraction_ground_loads,
            method,
            flow_type,
            load_years,
        )
        self.geometric_constraints = geometric_constraints
        # If a near-square design routine is requested, then we go from a
        # 1x1 to 32x32 at the B-spacing
        # The lower end of the near-square routine is always 1
        # There would never be a time that a user would __need__ to give a
        # different lower range. The upper number of boreholes range is
        # calculated based on the spacing and length provided.
        n = floor(self.geometric_constraints.length / self.geometric_constraints.b) + 1
        number_of_boreholes = int(n)
        self.coordinates_domain, self.fieldDescriptors = square_and_near_square(
            1, number_of_boreholes, self.geometric_constraints.b
        )

    def find_design(self, disp=False) -> Bisection1D:
        if disp:
            title = "Find near-square.."
            print(title + "\n" + len(title) * "=")
        return Bisection1D(
            self.coordinates_domain,
            self.fieldDescriptors,
            self.v_flow,
            self.borehole,
            self.fluid,
            self.pipe,
            self.grout,
            self.soil,
            self.max_boreholes,
            self.min_height,
            self.max_height,
            self.continue_if_design_unmet,
            self.start_month,
            self.end_month,
            self.min_EFT_allowable,
            self.max_EFT_allowable,
            self.hourly_extraction_ground_loads,
            method=self.method,
            flow_type=self.flow_type,
            disp=disp,
            field_type="near-square",
            load_years=self.load_years,
        )



from pygfunction.boreholes import Borehole

from ghedesigner.enums import FlowConfigType, TimestepType
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.ghe.search.bisection_1d import Bisection1D
from ghedesigner.media import Fluid, Grout, Soil


class Bisection2D(Bisection1D):
    def __init__(
        self,
        coordinates_domain_nested: list,
        field_descriptors: list,
        v_flow: float,
        borehole: Borehole,
        fluid: Fluid,
        pipe: Pipe,
        grout: Grout,
        soil: Soil,
        max_boreholes: int | None,
        min_height: float,
        max_height: float,
        continue_if_design_unmet: bool,
        start_month: int,
        end_month: int,
        min_eft: float,
        max_eft: float,
        hourly_extraction_ground_loads: list,
        method: TimestepType,
        flow_type: FlowConfigType = FlowConfigType.BOREHOLE,
        max_iter=15,
        disp=False,
        field_type="N/A",
        load_years=None,
    ) -> None:
        if load_years is None:
            load_years = [2019]
        if disp:
            print("Note: This routine requires a nested bisection search.")
        self.load_years = load_years
        # Get a coordinates domain for initialization
        coordinates_domain = coordinates_domain_nested[0]
        super().__init__(
            coordinates_domain,
            field_descriptors[0],
            v_flow,
            borehole,
            fluid,
            pipe,
            grout,
            soil,
            max_boreholes,
            min_height,
            max_height,
            continue_if_design_unmet,
            start_month,
            end_month,
            min_eft,
            max_eft,
            hourly_extraction_ground_loads,
            method=method,
            flow_type=flow_type,
            max_iter=max_iter,
            disp=disp,
            search=False,
            field_type=field_type,
            load_years=load_years,
        )

        # TODO why is the class variable set to an empty list and not the `coordinates_domain_nested` argument?
        # self.coordinates_domain_nested = []
        self.calculated_temperatures_nested = []
        # Tack on one borehole at the beginning to provide a high excess temperature
        outer_domain = [coordinates_domain_nested[0][0]]
        for cdn in coordinates_domain_nested:
            outer_domain.append(cdn[-1])

        self.coordinates_domain = outer_domain

        selection_key, _ = self.search()

        self.calculated_temperatures_nested.append(self.calculated_temperatures)

        # We tacked on one borehole to the beginning, so we need to subtract 1
        # on the index
        inner_domain = coordinates_domain_nested[selection_key - 1]
        self.coordinates_domain = inner_domain
        self.fieldDescriptors = field_descriptors[selection_key - 1]

        # Reset calculated temperatures
        self.calculated_temperatures = {}

        self.selection_key, self.selected_coordinates = self.search()



from pathlib import Path
from typing import Callable, cast

LoadFunctionType = Callable[[float], float]


class HeatPump:
    def __init__(self, name: str) -> None:
        self.name = name
        self.cop: float | None = None
        self.loads_list: list[float] | None = None
        self.load_function: LoadFunctionType | None = None

    def set_loads_from_file(self, file_path: Path) -> None:
        loads_file_content = file_path.read_text().strip()
        self.loads_list = [float(x) for x in loads_file_content.split()]
        self.load_function = self.get_load_from_list_at_time

    def set_loads_from_lambda(self, load_function: LoadFunctionType) -> None:
        self.load_function = load_function

    def get_load_from_list_at_time(self, time: float) -> float:
        if self.loads_list is None:
            raise ValueError("Loads values are not set.")
        # Assume time in hours? No leap year? No DST?
        hour_of_year = int(time % 8760)
        return self.loads_list[hour_of_year]

    def set_fixed_cop(self, cop: float) -> None:
        if cop <= 0:
            raise ValueError("Coefficient of Performance (COP) must be greater than zero.")
        self.cop = cop

    def do_sizing(self) -> None:
        pass

    def calculate(self, simulation_time_hours: float, loop_inlet_temp: float, flow_rate: float) -> float:
        if not self.load_function:
            raise ValueError("Load function is not set.")
        building_load = self.load_function(simulation_time_hours)
        if flow_rate == 0 or building_load == 0:
            return loop_inlet_temp

        # Check if cop is set to avoid division by zero
        if not self.cop:
            raise ValueError("Coefficient of Performance (COP) is not set.")

        if building_load > 0:
            # Heating the building, taking energy from loop
            loop_load = building_load - (building_load / self.cop)
        else:
            # Cooling the building, adding energy to loop
            loop_load = building_load + (building_load / self.cop)

        cp = 4100
        return loop_inlet_temp + loop_load / (flow_rate * cp)

    def get_ground_loads(self) -> list[float]:
        if self.loads_list is None:
            raise ValueError("Loads values are not set.")
        ground_loads: list[float] = []
        self.load_function = cast(LoadFunctionType, self.load_function)
        self.cop = cast(float, self.cop)
        for h in range(8760):
            building_load = self.load_function(h)
            if building_load > 0:
                # Heating the building, taking energy from loop
                loop_load = building_load - (building_load / self.cop)
            else:
                # Cooling the building, adding energy to loop
                loop_load = building_load + (building_load / self.cop)
            ground_loads.append(loop_load)
        return ground_loads



import logging
import warnings
from math import log

import numpy as np
import pygfunction as gt
from pygfunction.boreholes import Borehole
from pygfunction.enums import PipeType as PyPipeType
from pygfunction.gfunction import gFunction
from scipy.interpolate import interp1d, lagrange

from ghedesigner.enums import PipeType

logging.basicConfig(level=logging.WARN, format="%(message)s", datefmt="[%X]")
logger = logging.getLogger(__name__)

pyg_pipe_type_map = {
    PipeType.SINGLEUTUBE.name: PyPipeType.SINGLE_UTUBE.name,
    PipeType.DOUBLEUTUBESERIES.name: PyPipeType.DOUBLE_UTUBE_SERIES.name,
    PipeType.DOUBLEUTUBEPARALLEL.name: PyPipeType.DOUBLE_UTUBE_PARALLEL.name,
    PipeType.COAXIAL.name: PyPipeType.COAXIAL_ANNULAR_OUT.name,
}


def calculate_g_function(
    m_flow_borehole,
    bhe_type: PipeType,
    time_values,
    coordinates,
    borehole,
    fluid,
    pipe,
    grout,
    soil,
    boundary_condition="MIFT",
):
    match bhe_type:
        case PipeType.SINGLEUTUBE | PipeType.DOUBLEUTUBESERIES | PipeType.DOUBLEUTUBEPARALLEL:
            r_inner = pipe.r_in
            r_outer = pipe.r_out
        case PipeType.COAXIAL:
            # converting to pygfunction coaxial pipe conventions
            # this assumes the pipe (not annulus) is the inlet
            r_inner = [pipe.r_in[0], pipe.r_out[0]]
            r_outer = [pipe.r_in[1], pipe.r_out[1]]
        case _:
            raise ValueError(f"bhe_type {bhe_type} is not supported")

    # setup options
    # none of these were ever used or even exposed for users to access them. hardcoding them here until needed.
    solver = "equivalent"
    disp = False
    n_segments = 8
    end_length_ratio = 0.02

    options = {
        "nSegments": n_segments,
        "segment_ratios": gt.utilities.segment_ratios(nSegments=n_segments, end_length_ratio=end_length_ratio),
        "disp": disp,
    }

    nbh = len(coordinates)
    m_flow_network = nbh * m_flow_borehole

    g_func = gFunction.from_static_params(
        H=borehole.H,
        D=borehole.D,
        r_b=borehole.r_b,
        x=[x for x, _ in coordinates],
        y=[y for _, y in coordinates],
        alpha=soil.alpha,
        options=options,
        method=solver,
        boundary_condition=boundary_condition,
        k_p=pipe.k,
        k_s=soil.k,
        k_g=grout.k,
        epsilon=pipe.roughness,
        fluid_str=fluid.name,
        fluid_concentration_pct=fluid.concentration_percent,
        pos=pipe.pos,
        r_in=r_inner,
        r_out=r_outer,
        pipe_type_str=PyPipeType[pyg_pipe_type_map[bhe_type.name]].name,
        m_flow_network=m_flow_network,
    )

    g_func_vals = g_func.evaluate_g_function(time_values)

    return g_func_vals


def calc_g_func_for_multiple_lengths(
    b: float,
    h_values: list,
    r_b,
    depth,
    m_flow_borehole,
    bhe_type: PipeType,
    log_time,
    coordinates,
    fluid,
    pipe,
    grout,
    soil,
):
    r_b_values = dict.fromkeys(h_values, r_b)
    g_lts_values = {}

    alpha = soil.k / soil.rho_cp

    for h in h_values:
        borehole = Borehole(h, depth, r_b, 0.0, 0.0)

        ts = h**2 / (9.0 * alpha)  # Bore field characteristic time
        time_values = np.exp(log_time) * ts

        g_lts_values[h] = calculate_g_function(
            m_flow_borehole,
            bhe_type,
            time_values,
            coordinates,
            borehole,
            fluid,
            pipe,
            grout,
            soil,
        ).tolist()

    # Initialize the gFunction object
    return GFunction(
        b=b, r_b_values=r_b_values, d=depth, g_lts=g_lts_values, log_time=log_time, bore_locations=coordinates
    )


class GFunction:
    def __init__(
        self,
        b: float,
        d: float,
        r_b_values: dict,
        g_lts: dict,
        log_time: list,
        bore_locations: list,
    ) -> None:
        self.B: float = b  # a B spacing in the borefield
        # r_b (borehole radius) value keyed by height
        self.r_b_values: dict = r_b_values
        self.d: float = d  # burial depth
        self.g_lts: dict = g_lts  # g-functions (LTS) keyed by height
        # ln(t/ts) values that apply to all the heights
        self.log_time: list = log_time
        # (x, y) coordinates of boreholes
        self.bore_locations: list = bore_locations
        # self.time: dict = {}  # the time values in years

        # an interpolation table for B/H ratios, D, r_b (used in the method
        # g_function_interpolation)
        self.interpolation_table: dict = {}

    def g_function_interpolation(self, b_over_h: float, kind="default"):
        # the g-functions are stored in a dictionary based on heights, so an
        # equivalent height can be found
        h_eq = 1 / b_over_h * self.B

        tolerance = 0.001

        # Determine if we are out of range and need to extrapolate
        height_values = list(self.g_lts.keys())
        # If we are close to the outer bounds, then set H_eq as outer bounds
        close_tolerance = 1.0e-6
        if abs(h_eq - max(height_values)) < close_tolerance:
            h_eq = max(height_values)
        if abs(h_eq - min(height_values)) < close_tolerance:
            h_eq = min(height_values)

        if min(height_values) <= h_eq <= max(height_values) or abs(min(height_values) - h_eq) < tolerance:
            fill_value = ""
        else:
            fill_value = "extrapolate"
            warnings.warn("Extrapolation is being used.")

        # if the interpolation kind is default, use what we know about the
        # accuracy of interpolation to choose a technique

        if kind == "default":
            num_curves = len(height_values)
            if num_curves >= 5:  # noqa: PLR2004
                kind = "cubic"
            elif num_curves >= 3:  # noqa: PLR2004
                kind = "quadratic"
            elif num_curves == 2:  # noqa: PLR2004
                kind = "linear"
            elif (h_eq - height_values[0]) / height_values[0] < tolerance or min(height_values) - h_eq < tolerance:
                g_function = self.g_lts[height_values[0]]
                rb = self.r_b_values[height_values[0]]
                return g_function, rb, self.d, h_eq
            else:
                raise ValueError(
                    "The interpolation requires two g-function curves if the requested B/H is not already computed."
                )

        # Automatically adjust interpolation if necessary
        # Lagrange also needs 2
        interpolation_kinds = {"linear": 2, "quadratic": 3, "cubic": 4, "lagrange": 2}
        curves_by_kind = {2: "linear", 3: "quadratic", 4: "cubic"}
        if len(height_values) < 2:  # noqa: PLR2004
            raise ValueError("Interpolation requires two g-function curves.")
        else:
            # If the number of required curves for the interpolation type is
            # greater than what is available, reduce the interpolation type
            required_curves = interpolation_kinds[kind]
            if required_curves > len(height_values):
                kind = curves_by_kind[len(height_values)]

        # if the interpolation table is not yet know, build it
        if len(self.interpolation_table) == 0:
            # create an interpolation for the g-function which takes the height
            # (or equivalent height) as an input the g-function needs to be
            # interpolated at each point in dimensionless time
            self.interpolation_table["g"] = []
            for i, _ in enumerate(self.log_time):
                x = []
                y = []
                for key in self.g_lts:
                    height_value = float(key)
                    g_value = self.g_lts[key][i]
                    x.append(height_value)
                    y.append(g_value)
                f = lagrange(x, y) if kind == "lagrange" else interp1d(x, y, kind=kind, fill_value=fill_value)
                self.interpolation_table["g"].append(f)
            # create interpolation tables for 'D' and 'r_b' by height
            keys = list(self.r_b_values.keys())
            height_values_2 = []
            rb_values: list = []
            for h in keys:
                height_values_2.append(float(h))
                rb_values.append(self.r_b_values[h])
            if kind == "lagrange":
                rb_f = lagrange(height_values_2, rb_values)
            else:
                # interpolation function for rb values by H equivalent
                rb_f = interp1d(height_values_2, rb_values, kind=kind, fill_value=fill_value)
            self.interpolation_table["rb"] = rb_f

        # create the g-function by interpolating at each ln(t/ts) value
        rb_value = self.interpolation_table["rb"](h_eq)
        g_function_l = []
        for i in range(len(self.log_time)):
            f = self.interpolation_table["g"][i]
            g = f(h_eq).tolist()
            g_function_l.append(g)
        return g_function_l, rb_value, self.d, h_eq

    @staticmethod
    def borehole_radius_correction(g_function: list, rb: float, rb_star: float):
        r"""
        Correct the borehole radius. From paper 3 of Eskilson 1987.
        .. math::
            g(\dfrac{t}{t_s}, \dfrac{r_b^*}{H}) =
            g(\dfrac{t}{t_s}, \dfrac{r_b}{H}) - ln(\dfrac{r_b^*}{r_b})
        Parameters
        ----------
        g_function: list
            A g-function
        rb: float
            The current borehole radius
        rb_star: float
            The borehole radius that is being corrected to
        Returns
        -------
        g_function_corrected: list
            A corrected g_function
        """
        g_function_corrected = [g - log(rb_star / rb) for g in g_function]
        return g_function_corrected



#!/usr/bin/env python
import sys
from json import loads
from pathlib import Path

from jsonschema import validate
from jsonschema.exceptions import ValidationError

# Note: JSON schema does not currently have a good way to handle case-insensitive enums.
#       I think we should enforce case sensitivity.  The validation script should clearly alert the user.
#       More details here: https://github.com/json-schema-org/community/discussions/148


def validate_input_file(input_file_path: Path) -> None:
    """
    Validate input file against the schema
    """
    try:
        instance = loads(input_file_path.read_text())
        schema_path = Path(__file__).parent / "schemas" / "ghedesigner.schema.json"
        schema = loads(schema_path.read_text())
        validate(instance=instance, schema=schema)
    except ValidationError as error:
        print("\n Validation Error:", file=sys.stderr)
        print(f"  Bad Input Location: {' → '.join(map(str, error.path)) if error.path else 'Root'}", file=sys.stderr)
        print(f"  Problematic Key: {error.path[-1] if error.path else 'N/A'}", file=sys.stderr)
        print(f"  Error: {error.message}", file=sys.stderr)
        fix = "\n Suggested Fix:"
        if "required" in error.schema and isinstance(error.schema["required"], list):
            print(
                f"{fix}  Ensure that the following required keys are present: {', '.join(error.schema['required'])}",
                file=sys.stderr,
            )
        elif "enum" in error.schema:
            print(f"{fix}  Use one of the allowed values: {', '.join(map(str, error.schema['enum']))}", file=sys.stderr)
        elif "type" in error.schema:
            print(f"{fix}  Ensure the value is of type: {error.schema['type']}", file=sys.stderr)
        print("For example inputs, see the demo files.", file=sys.stderr)
        raise


if __name__ == "__main__":
    instance_path = Path(sys.argv[1]).resolve()
    try:
        validate_input_file(instance_path)
    except ValidationError:
        sys.exit(1)



from __future__ import annotations

import csv
from json import dumps, loads
from math import sqrt
from pathlib import Path
from typing import Any, Callable

import numpy as np
from scipy.interpolate import interp1d
from scipy.optimize import brentq


# Time functions
# --------------
def eskilson_log_times():
    # Return a list of Eskilson's original 27 dimensionless points in time
    return [
        -8.5,
        -7.8,
        -7.2,
        -6.5,
        -5.9,
        -5.2,
        -4.5,
        -3.963,
        -3.27,
        -2.864,
        -2.577,
        -2.171,
        -1.884,
        -1.191,
        -0.497,
        -0.274,
        -0.051,
        0.196,
        0.419,
        0.642,
        0.873,
        1.112,
        1.335,
        1.679,
        2.028,
        2.275,
        3.003,
    ]


# Spatial functions
# -----------------
def borehole_spacing(borehole, coordinates):
    # Use the distance between the first pair of coordinates as the B-spacing
    x_0, y_0 = coordinates[0]
    if len(coordinates) == 1:
        # Set the spacing to be the borehole radius if there's just one borehole
        return borehole.r_b
    elif len(coordinates) > 1:
        x_1, y_1 = coordinates[1]
        return max(borehole.r_b, sqrt((x_1 - x_0) ** 2 + (y_1 - y_0) ** 2))
    else:
        raise ValueError("The coordinates_domain needs to contain a positive number of (x, y) pairs.")


def length_of_side(n, b):
    return (n - 1) * b


# Design oriented functions
# -------------------------
def sign(x: float) -> int:
    """
    Determine the sign of a value, pronounced "sig-na"
    :param x: the input value
    :type x: float
    :return: a 1 or a -1
    """
    return int(abs(x) / x)


def check_bracket(sign_x_l, sign_x_r) -> bool:
    return sign_x_l < 0 < sign_x_r or sign_x_r < 0 < sign_x_l
    # True if bracketed the root


def solve_root(
    x: float,
    objective_function: Callable[..., Any],
    lower: float | None = None,
    upper: float | None = None,
    abs_tol=1.0e-6,
    rel_tol=1.0e-6,
    max_iter=50,
):
    # Vary flow rate to match the convective resistance

    # Use Brent Quadratic to find the root
    # Define a lower and upper for thermal conductivities
    if lower is None:
        lower = x / 100.0
    if upper is None:
        upper = x * 10.0
    # Check objective function upper and lower bounds to make sure the root is
    # bracketed
    minus = objective_function(lower)
    plus = objective_function(upper)
    # get signs of upper and lower thermal conductivity bounds
    kg_minus_sign = int(minus / abs(minus))
    kg_plus_sign = int(plus / abs(plus))

    # Solve the root if we can, if not, take the higher value
    if kg_plus_sign != kg_minus_sign:
        x = brentq(objective_function, lower, upper, xtol=abs_tol, rtol=rel_tol, maxiter=max_iter)
    elif kg_plus_sign == -1 and kg_minus_sign == -1:
        x = lower
    elif kg_plus_sign == 1 and kg_minus_sign == 1:
        x = upper

    return x


def write_idf_object(data: list) -> str:
    s = ""
    num_leading_pad_spaces = 4
    leading_pad = " " * num_leading_pad_spaces
    num_fields = len(data)

    for idx, (val, comment) in enumerate(data):
        len_val = len(val)
        comment_col_no = 30
        num_mid_pad = comment_col_no - num_leading_pad_spaces - len_val

        num_mid_pad = max(num_mid_pad, 2)

        mid_pad = " " * num_mid_pad

        if idx == 0:
            # writing object header
            s += f"{val},\n"
        elif (idx + 1) == num_fields:
            # writing last field. conclude with semicolon
            s += f"{leading_pad}{val};{mid_pad}!- {comment}\n"
        else:
            s += f"{leading_pad}{val},{mid_pad}!- {comment}\n"

    return s


def write_idf(summary_path: Path) -> None:
    data = loads(summary_path.read_text())

    # assuming the g-function file lives next to the summary file path...
    root_dir = summary_path.parent
    g_function_path = root_dir / "Gfunction.csv"
    g_function_arr = np.genfromtxt(g_function_path, delimiter=",")

    nbh = data["ghe_system"]["number_of_boreholes"]
    fluid_density = data["ghe_system"]["fluid_density"]["value"]
    fluid_mdot = data["ghe_system"]["fluid_mass_flow_rate_per_borehole"]["value"]
    des_vdot = fluid_mdot * nbh / fluid_density

    soil_k = data["ghe_system"]["soil_thermal_conductivity"]["value"]
    soil_rho_cp = data["ghe_system"]["soil_volumetric_heat_capacity"]["value"] * 1000

    system = [
        ("GroundHeatExchanger:System", ""),
        ("GHE System Name", "Name"),
        ("Inlet Node Name", "Inlet Node Name"),
        ("Outlet Node Name", "Outlet Node Name"),
        (f"{des_vdot:0.3e}", "Design Flow Rate {m3/s}"),
        ("Ground Temp Obj Type", "Undisturbed Ground Temperature Model Type"),
        ("Ground Temp Obj Name", "Undisturbed Ground Temperature Model Name"),
        (f"{soil_k:0.3f}", "Ground Thermal Conductivity {W/m-K}"),
        (f"{soil_rho_cp:0.3e}", "Ground Thermal Heat Capacity {J/m3-K}"),
        ("g-functions Obj Name", "GHE:Vertical:ResponseFactors Object Name"),
    ]

    bh_depth = data["ghe_system"]["borehole_buried_depth"]["value"]
    bh_length = data["ghe_system"]["active_borehole_length"]["value"]
    bh_dia = data["ghe_system"]["borehole_diameter"]["value"]
    grout_k = data["ghe_system"]["grout_thermal_conductivity"]["value"]
    grout_rho_cp = data["ghe_system"]["grout_volumetric_heat_capacity"]["value"] * 1000
    pipe_k = data["ghe_system"]["pipe_thermal_conductivity"]["value"]
    pipe_rho_cp = data["ghe_system"]["pipe_volumetric_heat_capacity"]["value"] * 1000
    pipe_outer_dia = data["ghe_system"]["pipe_geometry"]["pipe_outer_diameter"]["value"]
    pipe_inner_dia = data["ghe_system"]["pipe_geometry"]["pipe_inner_diameter"]["value"]
    pipe_thickness = (pipe_outer_dia - pipe_inner_dia) / 2.0
    shank_space = data["ghe_system"]["shank_spacing"]["value"] + pipe_outer_dia

    properties = [
        ("GroundHeatExchanger:Vertical:Properties", ""),
        ("Vert Props Name", "Name"),
        (f"{bh_depth:0.2f}", "Depth of Top of Borehole {m}"),
        (f"{bh_length:0.2f}", "Borehole Length {m}"),
        (f"{bh_dia:0.4f}", "Borehole Diameter {m}"),
        (f"{grout_k:0.2f}", "Grout Thermal Conductivity {W/m-K}"),
        (f"{grout_rho_cp:0.3e}", "Grout Thermal Heat Capacity {J/m3-K}"),
        (f"{pipe_k:0.2f}", "Pipe Thermal Conductivity {W/m-K}"),
        (f"{pipe_rho_cp:0.3e}", "Pipe Thermal Heat Capacity {J/m3-K}"),
        (f"{pipe_outer_dia:0.3e}", "Pipe Outer Diameter {m}"),
        (f"{pipe_thickness:0.3e}", "Pipe Thickness {m}"),
        (f"{shank_space:0.3e}", "U-Tube Distance {m}"),
    ]

    soil_density = 2500
    soil_cp = soil_rho_cp / soil_density
    ugt = data["ghe_system"]["soil_undisturbed_ground_temp"]["value"]

    ground_temps = [
        ("Site:GroundTemperature:Undisturbed:KusudaAchenbach", ""),
        ("GTM Name", "Name"),
        (f"{soil_k:0.2f}", "Soil Thermal Conductivity {W/m-K}"),
        (f"{soil_density:0.2f}", "Soil Density {kg/m3}"),
        (f"{soil_cp:0.2f}", "Soil Specific Heat {J/kg-K}"),
        (f"{ugt}", "Average Soil Surface Temperature {C}"),
        ("0", "Average Amplitude of Surface Temperature {deltaC}"),
        ("0", "Phase Shift of Minimum Surface Temperature {days}"),
    ]

    ref_ratio = (bh_dia / 2.0) / bh_length
    lntts_vals = g_function_arr[1:, 0]
    g_vals = g_function_arr[1:, 1]

    resp_factors = [
        ("GroundHeatExchanger:ResponseFactors", ""),
        ("Response Factors Name", "Name"),
        ("Vert Props Name", "GHE:Vertical:Properties Object Name"),
        (f"{int(nbh)}", "Number of Boreholes"),
        (f"{ref_ratio:0.2e}", "G-Function Reference Ratio {dimensionless}"),
    ]

    for idx in range(len(g_vals)):
        resp_factors.append((f"{lntts_vals[idx]:0.3f}", f"g-Function Ln(T/Ts) Value {idx + 1}"))
        resp_factors.append((f"{g_vals[idx]:0.3f}", f"g-Function g Value {idx + 1}"))

    s = ""
    s += write_idf_object(system)
    s += "\n"
    s += write_idf_object(properties)
    s += "\n"
    s += write_idf_object(ground_temps)
    s += "\n"
    s += write_idf_object(resp_factors)

    idf_path = root_dir / "out.idf"
    idf_path.write_text(s)


def check_arg_bounds(min_val: float, max_val: float, min_val_name: str, max_val_name: str) -> None:
    if min_val > max_val:
        # always throw error here
        raise ValueError(f"{min_val_name} ({min_val}) should be less than or equal to {max_val_name} ({max_val})")


def combine_sts_lts(log_time_lts: list, g_lts: list, log_time_sts: list, g_sts: list) -> interp1d:
    # make sure the short time step doesn't overlap with the long time step
    max_log_time_sts = max(log_time_sts)
    min_log_time_lts = min(log_time_lts)

    if max_log_time_sts < min_log_time_lts:
        log_time = log_time_sts + log_time_lts
        g = g_sts + g_lts
    else:
        # find where to stop in sts
        i = 0
        value = log_time_sts[i]
        while value <= min_log_time_lts:
            i += 1
            value = log_time_sts[i]
        log_time = log_time_sts[0:i] + log_time_lts
        g = g_sts[0:i] + g_lts
    g = interp1d(log_time, g)

    return g


def write_json(write_path: Path, input_dict: dict, indent: int = 2) -> None:
    with write_path.open("w") as f:
        f.write(dumps(input_dict, sort_keys=True, indent=indent, separators=(",", ": ")))


def write_flat_dict_to_csv(write_path: Path, input_dict: dict) -> None:
    with open(write_path, "w") as f:
        writer = csv.writer(f)
        writer.writerow(input_dict.keys())

        for row in zip(*input_dict.values()):
            writer.writerow(row)


def read_csv_column(file_path: str | Path, column: int | str, try_convert_to_numeric=True) -> list:
    """
    Reads a specific column from a CSV file.

    Parameters:
        file_path (str or Path): Path to the CSV file.
        column (str or int): Column name (str) or index (int) to extract.
        try_convert_to_numeric (bool, optional): If True, convert column to numeric. Defaults to True.

    Returns:
        list: A list of values from the specified column.
    """
    values = []

    with open(file_path, newline="", encoding="utf-8") as csv_file:
        reader = csv.reader(csv_file)
        header = next(reader)

        # Determine column index
        if isinstance(column, str):
            try:
                col_index = header.index(column)
            except ValueError:
                raise ValueError(f"Column name '{column}' not found in header.")
        elif isinstance(column, int):
            if column < 0 or column >= len(header):
                raise IndexError(f"Column index {column} out of range.")
            col_index = column
        else:
            raise TypeError("Column must be a string (name) or integer (index).")

        for row in reader:
            if len(row) > col_index:
                values.append(row[col_index])
            else:
                values.append("")  # Handle missing/short rows

        def try_convert(val):
            try:
                return float(val)
            except (ValueError, TypeError):
                return val

        if try_convert_to_numeric:
            values = [try_convert(x) for x in values]

    return values


def get_loads(loads_dict: dict) -> list[float]:
    if "load_values" in loads_dict:
        return loads_dict["load_values"]
    else:
        return read_csv_column(loads_dict["file_path"], loads_dict["column"])



from typing import cast

from bhr.borehole import Borehole as BHRBorehole
from pygfunction.boreholes import Borehole

from ghedesigner.constants import PI
from ghedesigner.ghe.boreholes.multi_u_borehole import GHEDesignerBoreholeWithMultiplePipes
from ghedesigner.ghe.boreholes.single_u_borehole import SingleUTube
from ghedesigner.ghe.pipe import Pipe
from ghedesigner.media import Fluid, Grout, Soil


class CoaxialPipe(GHEDesignerBoreholeWithMultiplePipes):
    def __init__(
        self, m_flow_borehole: float, fluid: Fluid, borehole: Borehole, pipe: Pipe, grout: Grout, soil: Soil
    ) -> None:
        super().__init__(m_flow_borehole, fluid, borehole, pipe, grout, soil)

        # Ensure r_in and r_out are list[float] rather than floats
        if not isinstance(pipe.r_out, list) or not isinstance(pipe.r_in, list):
            raise TypeError("pipe r_in and r_out must be list[float]")

        # Unpack the radii to reduce confusion in the future
        # Pipe naming nomenclature
        # <var>_<inner/outer pipe>_<inner/outer surface>
        # e.g. r_in_in is inner radius of the inner pipe
        self.r_inner = cast(tuple[float, float], pipe.r_in)
        self.r_outer = cast(tuple[float, float], pipe.r_out)
        self.r_in_in, self.r_in_out = self.r_inner
        self.r_out_in, self.r_out_out = self.r_outer

        # Store pipe roughness
        self.roughness = self.pipe.roughness

        self.bhr_borehole = BHRBorehole()
        self.bhr_borehole.init_coaxial_borehole(
            borehole_diameter=borehole.r_b * 2,
            outer_pipe_outer_diameter=(2.0 * self.r_out_out),
            outer_pipe_dimension_ratio=(2.0 * self.r_out_out) / (self.r_out_out - self.r_out_in),
            outer_pipe_conductivity=pipe.k[1],
            inner_pipe_outer_diameter=(2.0 * self.r_in_out),
            inner_pipe_dimension_ratio=(2.0 * self.r_in_out) / (self.r_in_out - self.r_in_in),
            inner_pipe_conductivity=pipe.k[0],
            length=borehole.H,
            grout_conductivity=grout.k,
            soil_conductivity=soil.k,
            fluid_type=self.fluid.name,
            fluid_concentration=self.fluid.concentration_percent / 100,
            boundary_condition="UNIFORM_BOREHOLE_WALL_TEMP",
        )

    def calc_effective_borehole_resistance(self) -> float:
        resist_bh_effective = self.bhr_borehole.calc_bh_resist(self.m_flow_borehole, self.soil.ugt)
        return resist_bh_effective

    def to_single(self) -> SingleUTube:
        # Find an equivalent single U-tube given a coaxial heat exchanger
        vol_fluid, vol_pipe = self.concentric_tube_volumes()

        resist_conv = self.bhr_borehole.calc_fluid_resist(self.m_flow_borehole, self.soil.ugt)
        resist_pipe = self.bhr_borehole.calc_pipe_cond_resist()
        preliminary = self.equivalent_single_u_tube(vol_fluid, vol_pipe, resist_conv, resist_pipe, self.pipe.rho_cp)

        # Vary grout thermal conductivity to match effective borehole thermal
        # resistance
        new_single_u_tube = self.match_effective_borehole_resistance(preliminary)

        return new_single_u_tube

    @staticmethod
    def compute_reynolds_concentric(m_flow_pipe: float, r_a_in: float, r_a_out: float, fluid: Fluid) -> float:
        # Hydraulic diameter and radius for concentric tube annulus region
        dia_hydraulic = 2 * (r_a_out - r_a_in)
        # r_h = dia_hydraulic / 2
        # Cross-sectional area of the annulus region
        area_cr_annular = PI * ((r_a_out**2) - (r_a_in**2))
        # Volume flow rate
        vol_flow_rate = m_flow_pipe / fluid.rho
        # Average velocity
        velocity = vol_flow_rate / area_cr_annular
        # Reynolds number
        return fluid.rho * velocity * dia_hydraulic / fluid.mu

    def as_dict(self) -> dict:
        reynold_no = self.compute_reynolds_concentric(self.m_flow_borehole, self.r_in_out, self.r_out_in, self.fluid)
        blob = {
            "type": str(self.__class__),
            "mass_flow_borehole": {"value": self.m_flow_borehole, "units": "kg/s"},
            "mass_flow_pipe": {"value": self.m_flow_borehole, "units": "kg/s"},
            "soil": self.soil.as_dict(),
            "grout": self.grout.as_dict(),
            "pipe": self.pipe.as_dict(),
            "reynolds": {"value": reynold_no, "units": ""},
        }
        return blob

    def concentric_tube_volumes(self) -> tuple[float, float]:
        # Compute volumes for concentric ghe geometry
        vol_fluid = PI * ((self.r_in_in**2) + (self.r_out_in**2) - (self.r_in_out**2))
        vol_pipe = PI * ((self.r_in_out**2) - (self.r_in_in**2) + (self.r_out_out**2) - (self.r_out_in**2))
        return vol_fluid, vol_pipe
